"""
Code for `sarif ls` command.
"""

from typing import List

from sarif import loader


def print_ls(files_or_dirs: List[str], output):
    """
    Print a SARIF file listing for each of the input files or directories.
    """
    dir_result = []
    for path in files_or_dirs:
        dir_result.append(f"{path}:")
        sarif_files = loader.load_sarif_files(path)
        if sarif_files:
            sarif_file_names = [f.get_file_name() for f in sarif_files]
            for file_name in sorted(sarif_file_names):
                dir_result.append(f"  {file_name}")
        else:
            dir_result.append("  (None)")
    if output:
        print("Writing file listing to", output)
        with open(output, "w", encoding="utf-8") as file_out:
            file_out.writelines(d + "\n" for d in dir_result)
    else:
        for directory in dir_result:
            print(directory)
        print()



"""
Top-level version information for sarif-tools.
"""
import importlib.metadata


def _read_package_version():
    try:
        return importlib.metadata.version("sarif-tools")
    except importlib.metadata.PackageNotFoundError:
        return "local"


__version__ = _read_package_version()



"""
Program entry point for sarif-tools on the command line.
"""

import argparse
from importlib import metadata
import os
import sys

from sarif import loader, sarif_file

from sarif.operations import (
    blame_op,
    copy_op,
    csv_op,
    diff_op,
    html_op,
    info_op,
    ls_op,
    summary_op,
    trend_op,
    word_op,
)


def main():
    """
    Entry point function.
    """
    args = ARG_PARSER.parse_args()

    if args.debug:
        print(f"SARIF tools v{SARIF_TOOLS_PACKAGE_VERSION}")
        print(f"Running code from {__file__}")
        known_args_summary = ", ".join(
            f"{key}={getattr(args, key)}" for key in vars(args)
        )
        print(f"Known arguments: {known_args_summary}")

    exitcode = args.func(args)
    return exitcode


def _create_arg_parser():
    cmd_list = "commands:\n"
    max_cmd_length = max(len(cmd) for cmd in _COMMANDS)
    col_width = max_cmd_length + 2
    for (cmd, cmd_attributes) in _COMMANDS.items():
        cmd_list += cmd.ljust(col_width) + cmd_attributes["desc"] + "\n"
    cmd_list += "Run `sarif <COMMAND> --help` for command-specific help."
    package_version = _read_package_version()
    parser = argparse.ArgumentParser(
        prog="sarif",
        description="Process sets of SARIF files",
        epilog=cmd_list,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(func=_usage)
    subparsers = parser.add_subparsers(dest="command", help="command")
    subparser = {}
    for (cmd, cmd_attributes) in _COMMANDS.items():
        subparser[cmd] = subparsers.add_parser(cmd, description=cmd_attributes["desc"])
        subparser[cmd].set_defaults(func=cmd_attributes["fn"])

    # Common options
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"sarif-tools version {package_version}",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print information useful for debugging"
    )
    parser.add_argument(
        "--check",
        "-x",
        type=str,
        choices=sarif_file.SARIF_SEVERITIES,
        help="Exit with error code if there are any issues of the specified level "
        + "(or for diff, an increase in issues at that level).",
    )

    for cmd in ["blame", "csv", "html", "summary", "word"]:
        subparser[cmd].add_argument(
            "--output", "-o", type=str, metavar="PATH", help="Output file or directory"
        )
    for cmd in ["copy", "diff", "info", "ls", "trend", "usage"]:
        subparser[cmd].add_argument(
            "--output", "-o", type=str, metavar="FILE", help="Output file"
        )

    for cmd in ["copy", "csv", "diff", "summary", "html", "trend", "word"]:
        subparser[cmd].add_argument(
            "--blame-filter",
            "-b",
            type=str,
            metavar="FILE",
            help="Specify the blame filter file to apply.  See README for format.",
        )

    # Command-specific options
    subparser["blame"].add_argument(
        "--code",
        "-c",
        metavar="PATH",
        type=str,
        help="Path to git repository; if not specified, the current working directory is used",
    )
    subparser["copy"].add_argument(
        "--timestamp",
        "-t",
        action="store_true",
        help='Append current timestamp to output filename in the "yyyymmddThhmmssZ" format used by '
        "the `sarif trend` command",
    )
    # csv defaults to no trimming
    subparser["csv"].add_argument(
        "--autotrim",
        "-a",
        action="store_true",
        help="Strip off the common prefix of paths in the CSV output",
    )
    # word and html default to trimming
    for cmd in ["html", "word"]:
        subparser[cmd].add_argument(
            "--no-autotrim",
            "-n",
            action="store_true",
            help="Do not strip off the common prefix of paths in the output document",
        )
        subparser[cmd].add_argument(
            "--image",
            type=str,
            help="Image to include at top of file - SARIF logo by default",
        )
    # csv, html and word allow trimmable paths to be specified
    for cmd in ["csv", "word", "html"]:
        subparser[cmd].add_argument(
            "--trim",
            metavar="PREFIX",
            action="append",
            type=str,
            help="Prefix to strip from issue paths, e.g. the checkout directory on the build agent",
        )
    # Most commands take an arbitrary list of SARIF files or directories
    for cmd in _COMMANDS:
        if cmd not in ["diff", "usage"]:
            subparser[cmd].add_argument(
                "files_or_dirs",
                metavar="file_or_dir",
                type=str,
                nargs="*",
                default=["."],
                help="A SARIF file or a directory containing SARIF files",
            )
    subparser["diff"].add_argument(
        "old_file_or_dir",
        type=str,
        nargs=1,
        help="An old SARIF file or a directory containing the old SARIF files",
    )
    subparser["diff"].add_argument(
        "new_file_or_dir",
        type=str,
        nargs=1,
        help="A new SARIF file or a directory containing the new SARIF files",
    )

    subparser["trend"].add_argument(
        "--dateformat",
        "-f",
        type=str,
        choices=["dmy", "mdy", "ymd"],
        default="dmy",
        help="Date component order to use in output CSV.  Default is `dmy`",
    )
    return parser


def _read_package_version():
    try:
        return metadata.version("sarif-tools")
    except metadata.PackageNotFoundError:
        return "local"


def _check(input_files: sarif_file.SarifFileSet, check_level):
    ret = 0
    if check_level:
        counts = input_files.get_result_count_by_severity()
        for severity in sarif_file.SARIF_SEVERITIES:
            ret += counts.get(severity, 0)
            if severity == check_level:
                break
    if ret > 0:
        sys.stderr.write(
            f"Check: exiting with return code {ret} due to issues at or above {check_level} "
            "severity\n"
        )
    return ret


def _load_blame_filter_file(file_path):
    filter_description = os.path.basename(file_path)
    include_substrings = []
    include_regexps = []
    exclude_substrings = []
    exclude_regexps = []
    try:
        with open(file_path, encoding="utf-8") as file_in:
            for line in file_in.readlines():
                if line.startswith("\ufeff"):
                    # Strip byte order mark
                    line = line[1:]
                lstrip = line.strip()
                if lstrip.startswith("#"):
                    # Ignore comment lines
                    continue
                pattern_spec = None
                is_include = True
                if lstrip.startswith("description:"):
                    filter_description = lstrip[12:].strip()
                elif lstrip.startswith("+: "):
                    is_include = True
                    pattern_spec = lstrip[3:].strip()
                elif lstrip.startswith("-: "):
                    is_include = False
                    pattern_spec = lstrip[3:].strip()
                else:
                    is_include = True
                    pattern_spec = lstrip
                if pattern_spec:
                    pattern_spec_len = len(pattern_spec)
                    if (
                        pattern_spec_len > 2
                        and pattern_spec.startswith("/")
                        and pattern_spec.endswith("/")
                    ):
                        (include_regexps if is_include else exclude_regexps).append(
                            pattern_spec[1 : pattern_spec_len - 1]
                        )
                    else:
                        (
                            include_substrings if is_include else exclude_substrings
                        ).append(pattern_spec)
    except UnicodeDecodeError as error:
        raise IOError(
            f"Cannot read blame filter file {file_path}: not UTF-8 encoded?"
        ) from error
    return (
        filter_description,
        include_substrings,
        include_regexps,
        exclude_substrings,
        exclude_regexps,
    )


def _init_blame_filtering(input_files, args):
    if args.blame_filter:
        filters = _load_blame_filter_file(args.blame_filter)
        input_files.init_blame_filter(*filters)


def _init_path_prefix_stripping(input_files, args, strip_by_default):
    if strip_by_default:
        autotrim = not args.no_autotrim
    else:
        autotrim = args.autotrim
    trim_paths = args.trim
    if autotrim or trim_paths:
        input_files.init_path_prefix_stripping(autotrim, trim_paths)


def _ensure_dir(dir_path):
    """
    Create directory if it does not exist
    """
    if dir_path and not os.path.isdir(dir_path):
        os.makedirs(dir_path)


def _prepare_output(
    input_files: sarif_file.SarifFileSet, output_arg, output_file_extension: str
):
    """
    Returns (output, output_multiple_files)
    output is args.output, or if that wasn't specified, a default output file based on the inputs
    and the file extension.
    output_multiple_files determines whether to output one file per input plus a totals file.
    It is false if there is only one input file, or args.output is a file that exists,
    or args.output ends with the expected file extension.
    """
    input_file_count = len(input_files)
    if input_file_count == 0:
        return ("static_analysis_output" + output_file_extension, False)
    if input_file_count == 1:
        derived_output_filename = (
            input_files[0].get_file_name_without_extension() + output_file_extension
        )
        if output_arg:
            if os.path.isdir(output_arg):
                return (os.path.join(output_arg, derived_output_filename), False)
            _ensure_dir(os.path.dirname(output_arg))
            return (output_arg, False)
        return (derived_output_filename, False)
    # Multiple input files
    if output_arg:
        if os.path.isfile(output_arg) or output_arg.strip().upper().endswith(
            output_file_extension.upper()
        ):
            # Output single file, even though there are multiple input files.
            _ensure_dir(os.path.dirname(output_arg))
            return (output_arg, False)
        _ensure_dir(output_arg)
        return (output_arg, True)
    return (os.getcwd(), True)


####################################### Command handlers #######################################


def _blame(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".sarif")
    blame_op.enhance_with_blame(
        input_files, args.code or os.getcwd(), output, multiple_file_output
    )
    return _check(input_files, args.check)


def _copy(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    _init_blame_filtering(input_files, args)
    output = args.output or "out.sarif"
    output_sarif_file_set = copy_op.generate_sarif(
        input_files,
        output,
        args.timestamp,
        SARIF_TOOLS_PACKAGE_VERSION,
        " ".join(sys.argv),
    )
    return _check(output_sarif_file_set, args.check)


def _csv(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=False)
    _init_blame_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".csv")
    csv_op.generate_csv(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _diff(args):
    original_sarif = loader.load_sarif_files(args.old_file_or_dir[0])
    new_sarif = loader.load_sarif_files(args.new_file_or_dir[0])
    _init_blame_filtering(original_sarif, args)
    _init_blame_filtering(new_sarif, args)
    return diff_op.print_diff(original_sarif, new_sarif, args.output, args.check)


def _html(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=True)
    _init_blame_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".html")
    html_op.generate_html(input_files, args.image, output, multiple_file_output)
    return _check(input_files, args.check)


def _info(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    info_op.generate_info(input_files, args.output)
    if args.check:
        return _check(input_files, args.check)
    return 0


def _ls(args):
    ls_op.print_ls(args.files_or_dirs, args.output)
    if args.check:
        input_files = loader.load_sarif_files(*args.files_or_dirs)
        return _check(input_files, args.check)
    return 0


def _summary(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    _init_blame_filtering(input_files, args)
    (output, multiple_file_output) = (None, False)
    if args.output:
        (output, multiple_file_output) = _prepare_output(
            input_files, args.output, ".txt"
        )
    summary_op.generate_summary(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _trend(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_blame_filtering(input_files, args)
    if args.output:
        _ensure_dir(os.path.dirname(args.output))
        output = args.output
    else:
        output = "static_analysis_trend.csv"
    trend_op.generate_trend_csv(input_files, output, args.dateformat)
    return _check(input_files, args.check)


def _usage(args):
    if hasattr(args, "output") and args.output:
        with open(args.output, "w", encoding="utf-8") as file_out:
            ARG_PARSER.print_help(file_out)
        print("Wrote usage instructions to", args.output)
    else:
        ARG_PARSER.print_help()
    if args.check:
        sys.stderr.write("Spurious --check argument")
        return 1
    return 0


def _word(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=True)
    _init_blame_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".docx")
    word_op.generate_word_docs_from_sarif_inputs(
        input_files, args.image, output, multiple_file_output
    )
    return _check(input_files, args.check)


_COMMANDS = {
    "blame": {
        "fn": _blame,
        "desc": "Enhance SARIF file with information from `git blame`",
    },
    "copy": {
        "fn": _copy,
        "desc": "Write a new SARIF file containing optionally-filtered data from other SARIF file(s)",
    },
    "csv": {
        "fn": _csv,
        "desc": "Write a CSV file listing the issues from the SARIF files(s) specified",
    },
    "diff": {
        "fn": _diff,
        "desc": "Find the difference between two [sets of] SARIF files",
    },
    "html": {
        "fn": _html,
        "desc": "Write an HTML representation of SARIF file(s) for viewing in a web browser",
    },
    "info": {"fn": _info, "desc": "Print information about SARIF file(s) structure"},
    "ls": {"fn": _ls, "desc": "List all SARIF files in the directories specified"},
    "summary": {
        "fn": _summary,
        "desc": "Write a text summary with the counts of issues from the SARIF files(s) specified",
    },
    "trend": {
        "fn": _trend,
        "desc": "Write a CSV file with time series data from SARIF files with "
        '"yyyymmddThhmmssZ" timestamps in their filenames',
    },
    "usage": {"fn": _usage, "desc": "(Command optional) - print usage and exit"},
    "word": {
        "fn": _word,
        "desc": "Produce MS Word .docx summaries of the SARIF files specified",
    },
}

SARIF_TOOLS_PACKAGE_VERSION = _read_package_version()
ARG_PARSER = _create_arg_parser()



"""
Code for `sarif copy` command.
"""

import copy
import datetime
import json
import os

from sarif import loader, sarif_file
from sarif.sarif_file import SarifFileSet


def generate_sarif(
    input_files: SarifFileSet,
    output: str,
    append_timestamp: bool,
    sarif_tools_version: str,
    cmdline: str,
):
    """
    Generate a new SARIF file based on the input files
    """
    sarif_data_out = {
        "$schema": "https://schemastore.azurewebsites.net/schemas/json/sarif-2.1.0-rtm.5.json",
        "version": "2.1.0",
        "runs": [],
    }
    now = datetime.datetime.utcnow()
    output_file_abs_path = os.path.abspath(output)
    conversion_timestamp_iso8601 = now.isoformat()
    conversion_timestamp_trendformat = now.strftime(sarif_file.DATETIME_FORMAT)
    run_count = 0
    input_file_count = 0
    for input_file in input_files:
        if input_file.get_abs_file_path() == output_file_abs_path:
            print(f"Auto-excluding output file {output} from input file list")
            continue
        input_file_count += 1
        input_file_path = input_file.get_abs_file_path()
        input_file_stat = os.stat(input_file_path)
        input_file_modified_iso8601 = datetime.datetime.fromtimestamp(
            input_file_stat.st_mtime
        ).isoformat()
        for input_run in input_file.runs:
            run_count += 1
            # Create a shallow copy
            input_run_json_copy = copy.copy(input_run.run_data)
            conversion_properties = {
                "file": input_file_path,
                "modified": input_file_modified_iso8601,
                "processed": conversion_timestamp_iso8601,
            }
            input_run_json_copy["conversion"] = {
                "tool": {
                    "driver": {
                        "name": "sarif-tools",
                        "fullName": "sarif-tools https://github.com/microsoft/sarif-tools/",
                        "version": sarif_tools_version,
                        "properties": conversion_properties,
                    }
                },
                "invocation": cmdline,
            }
            results = input_run.get_results()
            filter_stats = input_run.get_filter_stats()
            if filter_stats:
                input_run_json_copy["results"] = results
                conversion_properties["filtered"] = filter_stats.to_json_camel_case()
            sarif_data_out["runs"].append(input_run_json_copy)
    output_file_path = output
    if append_timestamp:
        output_split = os.path.splitext(output)
        output_file_path = (
            output_split[0]
            + f"_{conversion_timestamp_trendformat}"
            + (output_split[1] or ".sarif")
        )
    with open(output_file_path, "w", encoding="utf-8") as file_out:
        json.dump(sarif_data_out, file_out, indent=4)
    runs_string = "1 run" if run_count == 1 else f"{run_count} runs"
    files_string = (
        "1 SARIF file" if input_file_count == 1 else f"{input_file_count} SARIF files"
    )
    print(f"Wrote {output_file_path} with {runs_string} from {files_string}")
    total_filter_stats = input_files.get_filter_stats()
    if total_filter_stats:
        print(total_filter_stats.to_string())
    return loader.load_sarif_file(output_file_path)



"""
Defines classes representing sets of SARIF files, individual SARIF files and runs within SARIF
files, along with associated functions and constants.
"""

import copy
import datetime
import os
import re
from typing import Dict, Iterator, List, Optional, Tuple

SARIF_SEVERITIES = ["error", "warning", "note"]

RECORD_ATTRIBUTES = ["Tool", "Severity", "Code", "Location", "Line"]

# Standard time format for filenames, e.g. `20211012T110000Z` (not part of the SARIF standard).
# Can obtain from bash via `date +"%Y%m%dT%H%M%SZ"``
DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"
DATETIME_REGEX = r"\d{8}T\d{6}Z"

_SLASHES = ["\\", "/"]


def has_sarif_file_extension(filename):
    """
    As per section 3.2 of the SARIF standard, SARIF filenames SHOULD end in ".sarif" and MAY end in
    ".sarif.json".
    https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317421
    """
    filename_upper = filename.upper().strip()
    return any(filename_upper.endswith(x) for x in [".SARIF", ".SARIF.JSON"])


def _read_result_location(result) -> Tuple[str, str]:
    """
    Extract the file path and line number strings from the Result.
    Tools store this in different ways, so this function tries a few different JSON locations.
    """
    file_path = None
    line_number = None
    locations = result.get("locations", [])
    if locations:
        location = locations[0]
        physical_location = location.get("physicalLocation", {})
        # SpotBugs has some errors with no line number so deal with them by just leaving it at 1
        line_number = physical_location.get("region", {}).get("startLine", None)
        # For file name, first try the location written by DevSkim
        file_path = (
            location.get("physicalLocation", {})
            .get("address", {})
            .get("fullyQualifiedName", None)
        )
        if not file_path:
            # Next try the physical location written by MobSF and by SpotBugs (for some errors)
            file_path = (
                location.get("physicalLocation", {})
                .get("artifactLocation", {})
                .get("uri", None)
            )
        if not file_path:
            logical_locations = location.get("logicalLocations", None)
            if logical_locations:
                # Finally, try the logical location written by SpotBugs for some errors
                file_path = logical_locations[0].get("fullyQualifiedName", None)
    return (file_path, line_number)


def _group_records_by_severity(records) -> Dict[str, List[Dict]]:
    """
    Get the records, grouped by severity.
    """
    return {
        severity: [record for record in records if record["Severity"] == severity]
        for severity in SARIF_SEVERITIES
    }


def _count_records_by_issue_code(records, severity) -> List[Tuple]:
    """
    Return a list of pairs (code, count) of the records with the specified
    severities.
    """
    code_to_count = {}
    for record in records:
        if record["Severity"] == severity:
            code = record["Code"]
            code_to_count[code] = code_to_count.get(code, 0) + 1
    return sorted(code_to_count.items(), key=lambda x: x[1], reverse=True)


class FilterStats:
    """
    Statistics that record the outcome of a a filter.
    """

    def __init__(self, filter_description):
        self.filter_description = filter_description
        # Filter stats can also be loaded from a file created by `sarif copy`.
        self.rehydrated = False
        self.filter_datetime = None
        self.filtered_in_result_count = 0
        self.filtered_out_result_count = 0
        self.missing_blame_count = 0
        self.unconvincing_line_number_count = 0

    def reset_counters(self):
        """
        Zero all the counters.
        """
        self.filter_datetime = datetime.datetime.now()
        self.filtered_in_result_count = 0
        self.filtered_out_result_count = 0
        self.missing_blame_count = 0
        self.unconvincing_line_number_count = 0

    def add(self, other_filter_stats):
        """
        Add another set of filter stats to my totals.
        """
        if other_filter_stats:
            if other_filter_stats.filter_description and (
                other_filter_stats.filter_description != self.filter_description
            ):
                self.filter_description += f", {other_filter_stats.filter_description}"
            self.filtered_in_result_count += other_filter_stats.filtered_in_result_count
            self.filtered_out_result_count += (
                other_filter_stats.filtered_out_result_count
            )
            self.missing_blame_count += other_filter_stats.missing_blame_count
            self.unconvincing_line_number_count += (
                other_filter_stats.unconvincing_line_number_count
            )

    def __str__(self):
        """
        Automatic to_string()
        """
        return self.to_string()

    def to_string(self):
        """
        Generate a summary string for these filter stats.
        """
        ret = f"'{self.filter_description}'"
        if self.filter_datetime:
            ret += " at "
            ret += self.filter_datetime.strftime("%c")
        ret += (
            f": {self.filtered_out_result_count} filtered out, "
            f"{self.filtered_in_result_count} passed the filter"
        )
        if self.unconvincing_line_number_count:
            ret += (
                f", {self.unconvincing_line_number_count} included by default "
                "for lacking line number information"
            )
        if self.missing_blame_count:
            ret += (
                f", {self.missing_blame_count} included by default "
                "for lacking blame data to filter"
            )
        return ret

    def to_json_camel_case(self):
        """
        Generate filter stats as JSON using camelCase naming, to fit with SARIF standard section
        3.8.1 (Property Bags).
        """
        return {
            "filter": self.filter_description,
            "in": self.filtered_in_result_count,
            "out": self.filtered_out_result_count,
            "default": {
                "noLineNumber": self.unconvincing_line_number_count,
                "noBlame": self.missing_blame_count,
            },
        }


def load_filter_stats_from_json_camel_case(json_data):
    """
    Load filter stats from a SARIF file property bag
    """
    ret = None
    if json_data:
        ret = FilterStats(json_data["filter"])
        ret.rehydrated = True
        ret.filtered_in_result_count = json_data.get("in", 0)
        ret.filtered_out_result_count = json_data.get("out", 0)
        ret.unconvincing_line_number_count = json_data.get("default", {}).get(
            "noLineNumber", 0
        )
        ret.missing_blame_count = json_data.get("default", {}).get("noBlame", 0)
    return ret


def _add_filter_stats(accumulator, filter_stats):
    if filter_stats:
        if accumulator:
            accumulator.add(filter_stats)
            return accumulator
        return copy.copy(filter_stats)
    return accumulator


class _BlameFilter:
    """
    Class that implements blame filtering.
    """

    def __init__(self):
        self.filter_stats = None
        self.include_substrings = None
        self.include_regexes = None
        self.apply_inclusion_filter = False
        self.exclude_substrings = None
        self.exclude_regexes = None
        self.apply_exclusion_filter = False

    def init_blame_filter(
        self,
        filter_description,
        include_substrings,
        include_regexes,
        exclude_substrings,
        exclude_regexes,
    ):
        """
        Initialise the blame filter with the given filter patterns.
        """
        self.filter_stats = FilterStats(filter_description)
        self.include_substrings = (
            [s.upper().strip() for s in include_substrings]
            if include_substrings
            else None
        )
        self.include_regexes = include_regexes[:] if include_regexes else None
        self.apply_inclusion_filter = bool(
            self.include_substrings or self.include_regexes
        )
        self.exclude_substrings = (
            [s.upper().strip() for s in exclude_substrings]
            if exclude_substrings
            else None
        )
        self.exclude_regexes = exclude_regexes[:] if exclude_regexes else None
        self.apply_exclusion_filter = bool(
            self.exclude_substrings or self.exclude_regexes
        )

    def rehydrate_filter_stats(self, dehydrated_filter_stats, filter_datetime):
        """
        Restore filter stats from the SARIF file directly, where they were recordd when the filter
        was previously run.

        Note that if init_blame_filter is called, these rehydrated stats are discarded.
        """
        self.filter_stats = load_filter_stats_from_json_camel_case(
            dehydrated_filter_stats
        )
        self.filter_stats.filter_datetime = filter_datetime

    def _zero_counts(self):
        if self.filter_stats:
            self.filter_stats.reset_counters()

    def _check_include_result(self, author_mail):
        author_mail_upper = author_mail.upper().strip()
        matched_include_substrings = None
        matched_include_regexes = None
        if self.apply_inclusion_filter:
            if self.include_substrings:
                matched_include_substrings = [
                    s for s in self.include_substrings if s in author_mail_upper
                ]
            if self.include_regexes:
                matched_include_regexes = [
                    r
                    for r in self.include_regexes
                    if re.search(r, author_mail, re.IGNORECASE)
                ]
            if (not matched_include_substrings) and (not matched_include_regexes):
                return False
        if self.exclude_substrings and any(
            s in author_mail_upper for s in self.exclude_substrings
        ):
            return False
        if self.exclude_regexes and any(
            re.search(r, author_mail, re.IGNORECASE) for r in self.exclude_regexes
        ):
            return False
        return {
            "state": "included",
            "matchedSubstring": [s.lower() for s in matched_include_substrings]
            if matched_include_substrings
            else [],
            "matchedRegex": [r.lower() for r in matched_include_regexes]
            if matched_include_regexes
            else [],
        }

    def _filter_append(self, filtered_results, result, blame_info):
        # Remove any existing filter log on the result
        result.setdefault("properties", {}).pop("filtered", None)
        if blame_info:
            author_mail = blame_info.get("author-mail", None) or blame_info.get(
                "committer-mail", None
            )
            if author_mail:
                # First, check inclusion
                included = self._check_include_result(author_mail)
                if included:
                    self.filter_stats.filtered_in_result_count += 1
                    included["filter"] = self.filter_stats.filter_description
                    result["properties"]["filtered"] = included
                    filtered_results.append(result)
                else:
                    (_file_path, line_number) = _read_result_location(result)
                    if line_number == "1" or not line_number:
                        # Line number is not convincing.  Blame information may be misattributed.
                        self.filter_stats.unconvincing_line_number_count += 1
                        result["properties"]["filtered"] = {
                            "filter": self.filter_stats.filter_description,
                            "state": "default",
                            "missing": "line",
                        }
                        filtered_results.append(result)
                    else:
                        self.filter_stats.filtered_out_result_count += 1
            else:
                self.filter_stats.missing_blame_count += 1
                # Result did not contain complete blame information, so don't filter it out.
                result["properties"]["filtered"] = {
                    "filter": self.filter_stats.filter_description,
                    "state": "default",
                    "missing": "blame",
                }
                filtered_results.append(result)
        else:
            self.filter_stats.missing_blame_count += 1
            # Result did not contain blame information, so don't filter it out.
            filtered_results.append(result)

    def filter_results(self, results):
        """
        Apply this blame filter to a list of results, return the results that pass the filter
        and as a side-effect, update the filter stats.
        """
        if self.apply_inclusion_filter or self.apply_exclusion_filter:
            self._zero_counts()
            ret = []
            for result in results:
                blame_info = result.get("properties", {}).get("blame", None)
                self._filter_append(ret, result, blame_info)
            return ret
        # No inclusion or exclusion patterns
        return results

    def get_filter_stats(self) -> Optional[FilterStats]:
        """
        Get the statistics from running this filter.
        """
        return self.filter_stats


class SarifRun:
    """
    Class to hold a run object from a SARIF file (an entry in the top-level "runs" list
    in a SARIF file), as defined in SARIF standard section 3.14.
    https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317484
    """

    def __init__(self, sarif_file_object, run_index, run_data):
        self.sarif_file = sarif_file_object
        self.run_index = run_index
        self.run_data = run_data
        self._path_prefixes_upper = None
        self._cached_records = None
        self._filter = _BlameFilter()
        self._default_line_number = None
        conversion = run_data.get("conversion", None)
        if conversion:
            conversion_driver = conversion.get("tool", {}).get("driver", {})
            if conversion_driver.get("name", None) == "sarif-tools":
                # This run was written out by this tool!  Can restore filter stats.
                dehydrated_filter_stats = conversion_driver.get("properties", {}).get(
                    "filtered", None
                )
                if dehydrated_filter_stats:
                    filter_date = conversion_driver["properties"].get("processed", None)
                    self._filter.rehydrate_filter_stats(
                        dehydrated_filter_stats,
                        datetime.datetime.fromisoformat(filter_date)
                        if filter_date
                        else None,
                    )

    def init_path_prefix_stripping(self, autotrim=False, path_prefixes=None):
        """
        Set up path prefix stripping.  When records are subsequently obtained, the start of the
        path is stripped.
        If no path_prefixes are specified, the default behaviour is to strip the common prefix
        from each run.
        If path prefixes are specified, the specified prefixes are stripped.
        """
        prefixes = []
        if path_prefixes:
            prefixes = [prefix.strip().upper() for prefix in path_prefixes]
        if autotrim:
            autotrim_prefix = None
            records = self.get_records()
            if len(records) == 1:
                loc = records[0]["Location"].strip()
                slash_pos = max(loc.rfind(slash) for slash in _SLASHES)
                autotrim_prefix = loc[0:slash_pos] if slash_pos > -1 else None
            elif len(records) > 1:
                common_prefix = records[0]["Location"].strip()
                for record in records[1:]:
                    for (char_pos, char) in enumerate(record["Location"].strip()):
                        if char_pos >= len(common_prefix):
                            break
                        if char != common_prefix[char_pos]:
                            common_prefix = common_prefix[0:char_pos]
                            break
                    if not common_prefix:
                        break
                if common_prefix:
                    autotrim_prefix = common_prefix.upper()
            if autotrim_prefix and not any(
                p.startswith(autotrim_prefix.strip().upper()) for p in prefixes
            ):
                prefixes.append(autotrim_prefix)
        self._path_prefixes_upper = prefixes or None
        # Clear the untrimmed records cached by get_records() above.
        self._cached_records = None

    def init_default_line_number_1(self):
        """
        Some SARIF records lack a line number.  If this method is called, the default line number
        "1" is substituted in that case in the records returned by get_records().  Otherwise,
        None is returned.
        """
        self._default_line_number = "1"
        self._cached_records = None

    def init_blame_filter(
        self,
        filter_description,
        include_substrings,
        include_regexes,
        exclude_substrings,
        exclude_regexes,
    ):
        """
        Set up blame filtering.  This is applied to the author_mail field added to the "blame"
        property bag in each SARIF file.  Raises an error if any of the SARIF files don't contain
        blame information.
        If only inclusion criteria are provided, only issues matching the inclusion criteria
        are considered.
        If only exclusion criteria are provided, only issues not matching the exclusion criteria
        are considered.
        If both are provided, only issues matching the inclusion criteria and not matching the
        exclusion criteria are considered.
        include_substrings = substrings of author_mail to filter issues for inclusion.
        include_regexes = regular expressions for author_mail to filter issues for inclusion.
        exclude_substrings = substrings of author_mail to filter issues for exclusion.
        exclude_regexes = regular expressions for author_mail to filter issues for exclusion.
        """
        self._filter.init_blame_filter(
            filter_description,
            include_substrings,
            include_regexes,
            exclude_substrings,
            exclude_regexes,
        )
        # Clear the unfiltered records cached by get_records() above.
        self._cached_records = None

    def get_tool_name(self) -> str:
        """
        Get the tool name from this run.
        """
        return self.run_data["tool"]["driver"]["name"]

    def get_conversion_tool_name(self) -> Optional[str]:
        """
        Get the conversion tool name from this run, if any.
        """
        if "conversion" in self.run_data:
            return (
                self.run_data["conversion"]["tool"].get("driver", {}).get("name", None)
            )
        return None

    def get_results(self) -> List[Dict]:
        """
        Get the results from this run.  These are the Result objects as defined in the SARIF
        standard section 3.27.  The results are filtered if a filter has ben configured.
        https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317638
        """
        return self._filter.filter_results(self.run_data["results"])

    def get_records(self) -> List[Dict]:
        """
        Get simplified records derived from the results of this run.  The records have the
        keys defined in `RECORD_ATTRIBUTES`.
        """
        if not self._cached_records:
            results = self.get_results()
            self._cached_records = [self.result_to_record(result) for result in results]
        return self._cached_records

    def get_records_grouped_by_severity(self) -> Dict[str, List[Dict]]:
        """
        Get the records, grouped by severity.
        """
        return _group_records_by_severity(self.get_records())

    def result_to_record(self, result):
        """
        Convert a SARIF result object to a simple record with fields "Tool", "Location", "Line",
        "Severity" and "Code".
        See definition of result object here:
        https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317638
        """
        error_id = result["ruleId"]
        tool_name = self.get_tool_name()
        (file_path, line_number) = _read_result_location(result)
        if not file_path:
            raise ValueError(f"No location in {error_id} output from {tool_name}")
        if not line_number:
            line_number = "1"

        if self._path_prefixes_upper:
            file_path_upper = file_path.upper()
            for prefix in self._path_prefixes_upper:
                if file_path_upper.startswith(prefix):
                    prefixlen = len(prefix)
                    if len(file_path) > prefixlen and file_path[prefixlen] in _SLASHES:
                        # Strip off trailing path separator
                        file_path = file_path[prefixlen + 1 :]
                    else:
                        file_path = file_path[prefixlen:]
                    break

        # Get the error severity, if included, and code
        severity = result.get(
            "level", "warning"
        )  # If an error has no specified level then by default it is a warning
        message = result["message"]["text"]

        # Create a dict representing this result
        record = {
            "Tool": tool_name,
            "Location": file_path,
            "Line": line_number,
            "Severity": severity,
            "Code": f"{error_id} {message}",
        }
        return record

    def get_result_count(self) -> int:
        """
        Return the total number of results.
        """
        return len(self.get_results())

    def get_result_count_by_severity(self) -> Dict[str, int]:
        """
        Return a dict from SARIF severity to number of records.
        """
        records = self.get_records()
        return {
            severity: sum(1 for record in records if severity in record["Severity"])
            for severity in SARIF_SEVERITIES
        }

    def get_issue_code_histogram(self, severity) -> List[Tuple]:
        """
        Return a list of pairs (code, count) of the records with the specified
        severities.
        """
        return _count_records_by_issue_code(self.get_records(), severity)

    def get_filter_stats(self) -> Optional[FilterStats]:
        """
        Get the number of records that were included or excluded by the filter.
        """
        return self._filter.get_filter_stats()


class SarifFile:
    """
    Class to hold SARIF data parsed from a file and provide accesssors to the data.
    """

    def __init__(self, file_path, data):
        self.abs_file_path = os.path.abspath(file_path)
        self.data = data
        self.runs = [
            SarifRun(self, run_index, run_data)
            for (run_index, run_data) in enumerate(self.data.get("runs", []))
        ]

    def __bool__(self):
        """
        True if non-empty.
        """
        return bool(self.runs)

    def init_path_prefix_stripping(self, autotrim=False, path_prefixes=None):
        """
        Set up path prefix stripping.  When records are subsequently obtained, the start of the
        path is stripped.
        If no path_prefixes are specified, the default behaviour is to strip the common prefix
        from each run.
        If path prefixes are specified, the specified prefixes are stripped.
        """
        for run in self.runs:
            run.init_path_prefix_stripping(autotrim, path_prefixes)

    def init_default_line_number_1(self):
        """
        Some SARIF records lack a line number.  If this method is called, the default line number
        "1" is substituted in that case in the records returned by get_records().  Otherwise,
        None is returned.
        """
        for run in self.runs:
            run.init_default_line_number_1()

    def init_blame_filter(
        self,
        filter_description,
        include_substrings,
        include_regexes,
        exclude_substrings,
        exclude_regexes,
    ):
        """
        Set up blame filtering.  This is applied to the author_mail field added to the "blame"
        property bag in each SARIF file.  Raises an error if any of the SARIF files don't contain
        blame information.
        If only inclusion criteria are provided, only issues matching the inclusion criteria
        are considered.
        If only exclusion criteria are provided, only issues not matching the exclusion criteria
        are considered.
        If both are provided, only issues matching the inclusion criteria and not matching the
        exclusion criteria are considered.
        include_substrings = substrings of author_mail to filter issues for inclusion.
        include_regexes = regular expressions for author_mail to filter issues for inclusion.
        exclude_substrings = substrings of author_mail to filter issues for exclusion.
        exclude_regexes = regular expressions for author_mail to filter issues for exclusion.
        """
        for run in self.runs:
            run.init_blame_filter(
                filter_description,
                include_substrings,
                include_regexes,
                exclude_substrings,
                exclude_regexes,
            )

    def get_abs_file_path(self) -> str:
        """
        Get the absolute file path from which this SARIF data was loaded.
        """
        return self.abs_file_path

    def get_file_name(self) -> str:
        """
        Get the file name from which this SARIF data was loaded.
        """
        return os.path.basename(self.abs_file_path)

    def get_file_name_without_extension(self) -> str:
        """
        Get the file name from which this SARIF data was loaded, without extension.
        """
        file_name = self.get_file_name()
        return file_name[0 : file_name.index(".")] if "." in file_name else file_name

    def get_file_name_extension(self) -> str:
        """
        Get the extension of the file name from which this SARIF data was loaded.
        Initial "." exlcuded.
        """
        file_name = self.get_file_name()
        return file_name[file_name.index(".") + 1 :] if "." in file_name else ""

    def get_filename_timestamp(self) -> str:
        """
        Extract the timestamp from the filename and return the date-time string extracted.
        """
        parsed_date = re.findall(DATETIME_REGEX, self.get_file_name())
        return parsed_date if len(parsed_date) == 1 else None

    def get_distinct_tool_names(self):
        """
        Return a list of tool names that feature in the runs in this file.
        The list is deduplicated and sorted into alphabetical order.
        """
        return sorted(list(set(run.get_tool_name() for run in self.runs)))

    def get_results(self) -> List[Dict]:
        """
        Get the results from all runs in this file.  These are the Result objects as defined in the
        SARIF standard section 3.27.
        https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317638
        """
        ret = []
        for run in self.runs:
            ret += run.get_results()
        return ret

    def get_records(self) -> List[Dict]:
        """
        Get simplified records derived from the results of all runs.  The records have the
        keys defined in `RECORD_ATTRIBUTES`.
        """
        ret = []
        for run in self.runs:
            ret += run.get_records()
        return ret

    def get_records_grouped_by_severity(self) -> Dict[str, List[Dict]]:
        """
        Get the records, grouped by severity.
        """
        return _group_records_by_severity(self.get_records())

    def get_result_count(self) -> int:
        """
        Return the total number of results.
        """
        return sum(run.get_result_count() for run in self.runs)

    def get_result_count_by_severity(self) -> Dict[str, int]:
        """
        Return a dict from SARIF severity to number of records.
        """
        get_result_count_by_severity_per_run = [
            run.get_result_count_by_severity() for run in self.runs
        ]
        return {
            severity: sum(
                rc.get(severity, 0) for rc in get_result_count_by_severity_per_run
            )
            for severity in SARIF_SEVERITIES
        }

    def get_issue_code_histogram(self, severity) -> List[Tuple]:
        """
        Return a list of pairs (code, count) of the records with the specified
        severities.
        """
        return _count_records_by_issue_code(self.get_records(), severity)

    def get_filter_stats(self) -> Optional[FilterStats]:
        """
        Get the number of records that were included or excluded by the filter.
        """
        ret = None
        for run in self.runs:
            ret = _add_filter_stats(ret, run.get_filter_stats())
        return ret


class SarifFileSet:
    """
    Class representing a set of SARIF files.
    The "composite" pattern is used to allow multiple subdirectories.
    """

    def __init__(self):
        self.subdirs = []
        self.files = []

    def __bool__(self):
        """
        Return true if there are any SARIF files, regardless of whether they contain any runs.
        """
        return any(bool(subdir) for subdir in self.subdirs) or bool(self.files)

    def __len__(self):
        """
        Return the number of SARIF files, in total.
        """
        return sum(len(subdir) for subdir in self.subdirs) + sum(
            1 for f in self.files if f
        )

    def __iter__(self) -> Iterator[SarifFile]:
        """
        Iterate the SARIF files in this set.
        """
        for subdir in self.subdirs:
            for input_file in subdir.files:
                yield input_file
        for input_file in self.files:
            yield input_file

    def __getitem__(self, index) -> SarifFile:
        i = 0
        for subdir in self.subdirs:
            for input_file in subdir.files:
                if i == index:
                    return input_file
                i += 1
        return self.files[index - i]

    def get_description(self):
        """
        Get a description of the SARIF file set - the name of the single file or the number of
        files.
        """
        count = len(self)
        if count == 1:
            return self[0].get_file_name()
        return f"{count} files"

    def init_path_prefix_stripping(self, autotrim=False, path_prefixes=None):
        """
        Set up path prefix stripping.  When records are subsequently obtained, the start of the
        path is stripped.
        If no path_prefixes are specified, the default behaviour is to strip the common prefix
        from each run.
        If path prefixes are specified, the specified prefixes are stripped.
        """
        for subdir in self.subdirs:
            subdir.init_path_prefix_stripping(autotrim, path_prefixes)
        for input_file in self.files:
            input_file.init_path_prefix_stripping(autotrim, path_prefixes)

    def init_default_line_number_1(self):
        """
        Some SARIF records lack a line number.  If this method is called, the default line number
        "1" is substituted in that case in the records returned by get_records().  Otherwise,
        None is returned.
        """
        for subdir in self.subdirs:
            subdir.init_default_line_number_1()
        for input_file in self.files:
            input_file.init_default_line_number_1()

    def init_blame_filter(
        self,
        filter_description,
        include_substrings,
        include_regexes,
        exclude_substrings,
        exclude_regexes,
    ):
        """
        Set up blame filtering.  This is applied to the author_mail field added to the "blame"
        property bag in each SARIF file.  Raises an error if any of the SARIF files don't contain
        blame information.
        If only inclusion criteria are provided, only issues matching the inclusion criteria
        are considered.
        If only exclusion criteria are provided, only issues not matching the exclusion criteria
        are considered.
        If both are provided, only issues matching the inclusion criteria and not matching the
        exclusion criteria are considered.
        include_substrings = substrings of author_mail to filter issues for inclusion.
        include_regexes = regular expressions for author_mail to filter issues for inclusion.
        exclude_substrings = substrings of author_mail to filter issues for exclusion.
        exclude_regexes = regular expressions for author_mail to filter issues for exclusion.
        """
        for subdir in self.subdirs:
            subdir.init_blame_filter(
                filter_description,
                include_substrings,
                include_regexes,
                exclude_substrings,
                exclude_regexes,
            )
        for input_file in self.files:
            input_file.init_blame_filter(
                filter_description,
                include_substrings,
                include_regexes,
                exclude_substrings,
                exclude_regexes,
            )

    def add_dir(self, sarif_file_set):
        """
        Add a SarifFileSet as a subdirectory.
        """
        self.subdirs.append(sarif_file_set)

    def add_file(self, sarif_file_object: SarifFile):
        """
        Add a single SARIF file to the set.
        """
        self.files.append(sarif_file_object)

    def get_distinct_tool_names(self) -> List[str]:
        """
        Return a list of tool names that feature in the runs in these files.
        The list is deduplicated and sorted into alphabetical order.
        """
        all_tool_names = set()
        for subdir in self.subdirs:
            all_tool_names.update(subdir.get_distinct_tool_names())
        for input_file in self.files:
            all_tool_names.update(input_file.get_distinct_tool_names())

        return sorted(list(all_tool_names))

    def get_results(self) -> List[Dict]:
        """
        Get the results from all runs in all files.  These are the Result objects as defined in the
        SARIF standard section 3.27.
        https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317638
        """
        ret = []
        for subdir in self.subdirs:
            ret += subdir.get_results()
        for input_file in self.files:
            ret += input_file.get_results()
        return ret

    def get_records(self) -> List[Dict]:
        """
        Get simplified records derived from the results of all runs.  The records have the
        keys defined in `RECORD_ATTRIBUTES`.
        """
        ret = []
        for subdir in self.subdirs:
            ret += subdir.get_records()
        for input_file in self.files:
            ret += input_file.get_records()
        return ret

    def get_records_grouped_by_severity(self) -> Dict[str, List[Dict]]:
        """
        Get the records, grouped by severity.
        """
        return _group_records_by_severity(self.get_records())

    def get_result_count(self) -> int:
        """
        Return the total number of results.
        """
        return sum(subdir.get_result_count() for subdir in self.subdirs) + sum(
            input_file.get_result_count() for input_file in self.files
        )

    def get_result_count_by_severity(self) -> Dict[str, int]:
        """
        Return a dict from SARIF severity to number of records.
        """
        result_counts_by_severity = []
        for subdir in self.subdirs:
            result_counts_by_severity.append(subdir.get_result_count_by_severity())
        for input_file in self.files:
            result_counts_by_severity.append(input_file.get_result_count_by_severity())
        return {
            severity: sum(rc.get(severity, 0) for rc in result_counts_by_severity)
            for severity in SARIF_SEVERITIES
        }

    def get_issue_code_histogram(self, severity) -> List[Tuple]:
        """
        Return a list of pairs (code, count) of the records with the specified
        severities.
        """
        return _count_records_by_issue_code(self.get_records(), severity)

    def get_filter_stats(self) -> Optional[FilterStats]:
        """
        Get the number of records that were included or excluded by the filter.
        """
        ret = None
        for subdir in self.subdirs:
            ret = _add_filter_stats(ret, subdir.get_filter_stats())
        for input_file in self.files:
            ret = _add_filter_stats(ret, input_file.get_filter_stats())
        return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sarif', 'sarif.cmdline', 'sarif.operations']

package_data = \
{'': ['*'], 'sarif.operations': ['templates/*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'python-docx>=0.8.11,<0.9.0']

entry_points = \
{'console_scripts': ['sarif = sarif.cmdline.main:main']}

setup_kwargs = {
    'name': 'sarif-tools',
    'version': '1.0.0',
    'description': 'SARIF tools',
    'long_description': '# SARIF Tools\n\nA set of command line tools and Python library for working with SARIF files.\n\nRead more about the SARIF format here: https://sarifweb.azurewebsites.net/\n\n# Installation\n\n## Prerequisites\n\nYou need Python 3.8 or later installed.  Get it from [python.org](https://www.python.org/downloads/).  This document assumes that the `python` command runs that version.\n\n## Installing on Windows\n\nOpen an Admin Command Prompt (Start > Command Prompt > Run as Administrator) and type:\n```\npip install sarif-tools\n```\n\n## Installing on Linux or Mac\n\n```\nsudo pip install sarif-tools\n```\n\n## Testing the installation\n\nAfter installing using `pip`, you should then be able to run:\n```\nsarif --version\n```\n\n## Troubleshooting installation\n\nThis section has suggestions in case the `sarif` command is not available after installation.\n\nA launcher called `sarif` or `sarif.exe` is created in the Python installation\'s `Scripts` directory.  The `Scripts` directory needs to be in the `PATH`\nenvironment variable for you to be able to type `sarif` at the command prompt; this is most likely the case if `pip` is run as a\nsuper-user when installing (e.g. Administrator Command Prompt on Windows, or using `sudo` on Linux).\n\nIf the `Scripts` directory is not in the `PATH`, then you need to type `python -m sarif` instead of `sarif` to run the tool.\n\nConfusion can arise when the `python` and `pip` commands on the `PATH` are from different installations, or the `python` installation on the super-user\'s `PATH` is different from the `python` command on the normal user\'s path.  On Windows, you can use `where python` and `where pip` in normal CMD and Admin CMD to see which installations are in use; on Linux, it\'s `which python` and `which pip` with and without `sudo`.\n\n# Command Line Usage\n\n```\nusage: sarif [-h] [--version] [--debug] [--check {error,warning,note}] {blame,copy,csv,diff,html,info,ls,summary,trend,usage,word} ...\n\nProcess sets of SARIF files\n\npositional arguments:\n  {blame,copy,csv,diff,html,info,ls,summary,trend,usage,word}\n                        command\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --version, -v         show program\'s version number and exit\n  --debug               Print information useful for debugging\n  --check {error,warning,note}, -x {error,warning,note}\n                        Exit with error code if there are any issues of the specified level (or for diff, an increase in issues at that level).\n\ncommands:\nblame    Enhance SARIF file with information from `git blame`\ncopy     Write a new SARIF file containing optionally-filtered data from other SARIF file(s)\ncsv      Write a CSV file listing the issues from the SARIF files(s) specified\ndiff     Find the difference between two [sets of] SARIF files\nhtml     Write an HTML representation of SARIF file(s) for viewing in a web browser\ninfo     Print information about SARIF file(s) structure\nls       List all SARIF files in the directories specified\nsummary  Write a text summary with the counts of issues from the SARIF files(s) specified\ntrend    Write a CSV file with time series data from SARIF files with "yyyymmddThhmmssZ" timestamps in their filenames\nusage    (Command optional) - print usage and exit\nword     Produce MS Word .docx summaries of the SARIF files specified\nRun `sarif <COMMAND> --help` for command-specific help.\n```\n\n## Commands\n\nThe commands are illustrated below assuming input files in the following locations:\n\n- `C:\\temp\\sarif_files` = a directory of SARIF files with arbitrary filenames.\n- `C:\\temp\\sarif_with_date` = a directory of SARIF files with filenames including timestamps e.g. `C:\\temp\\sarif_with_date\\myapp_devskim_output_20211001T012000Z.sarif`.\n- `C:\\temp\\old_sarif_files` = a directory of SARIF files with arbitrary filenames from an older build.\n- `C:\\code\\my_source_repo` = checkout directory of source code files from which SARIF results were obtained.\n\n### blame\n```\nusage: sarif blame [-h] [--output PATH] [--code PATH] [file_or_dir [file_or_dir ...]]\n\nEnhance SARIF file with information from `git blame`\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output PATH, -o PATH\n                        Output file or directory\n  --code PATH, -c PATH  Path to git repository; if not specified, the current working directory is used\n```\n\nAugment SARIF files with `git blame` information, and write the augmented files to a specified location.\n```shell\nsarif blame -o "C:\\temp\\sarif_files_with_blame_info" -c "C:\\code\\my_source_repo" "C:\\temp\\sarif_files"\n```\n\nIf the current working directory is the git repository, the `-c` argument can be omitted.\n\nSee [Blame filtering](blame-filtering) below for the format of the blame information that gets added to the SARIF files.\n\n### copy\n\n```\nusage: sarif copy [-h] [--output FILE] [--blame-filter FILE] [--timestamp] [file_or_dir [file_or_dir ...]]\n\nWrite a new SARIF file containing optionally-filtered data from other SARIF file(s)\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n  --timestamp, -t       Append current timestamp to output filename in the "yyyymmddThhmmssZ" format used by the `sarif trend` command\n```\n\nWrite a new SARIF file containing optionally-filtered data from an existing SARIF file or multiple\nSARIF files.  The resulting file contains each run from the original SARIF files back-to-back.\nThe results can be filtered (see [Blame filtering](blame-filtering) below), in which case only\nthose results from the original SARIF files that meet the filter are included; the output file\ncontains no information about the excluded records.  If a run in the original file was empty,\nor all its results are filtered out, the empty run is still included.\n\nIf no output filename is provided, a file called `out.sarif` in the current directory is written.\nIf the output file already exists and is also in the input file list, it is not included in the\ninputs, to avoid duplication of results.  The output file is overwritten without warning.\n\nThe `file_or_dir` specifier can include wildcards e.g. `c:\\temp\\**\\devskim*.sarif` (i.e.\na "glob").  This works for all commands, but it is particularly useful for `copy`.\n\nOne use for this is to combine a set of SARIF files from multiple static analysis tools run during\na build process into a single file that can be more easily stored and processed as a build asset.\n\n### csv\n\n```\nusage: sarif csv [-h] [--output PATH] [--blame-filter FILE] [--autotrim] [--trim PREFIX] [file_or_dir [file_or_dir ...]]\n\nWrite a CSV file listing the issues from the SARIF files(s) specified\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output PATH, -o PATH\n                        Output file or directory\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n  --autotrim, -a        Strip off the common prefix of paths in the CSV output\n  --trim PREFIX         Prefix to strip from issue paths, e.g. the checkout directory on the build agent\n```\n\nWrite out a simple tabular list of issues from [a set of] SARIF files.  This can then be analysed, e.g. via Pivot Tables in Excel.\n\nUse the `--trim` option to strip specific prefixes from the paths, to make the CSV less verbose.  Alternatively, use `--autotrim` to strip off the longest common prefix.\n\nGenerate a CSV summary of a single SARIF file with common file path prefix suppressed:\n```shell\nsarif csv "C:\\temp\\sarif_files\\devskim_myapp.sarif"\n```\n\nGenerate a CSV summary of a directory of SARIF files with path prefix `C:\\code\\my_source_repo` suppressed:\n```shell\nsarif csv --trim c:\\code\\my_source_repo "C:\\temp\\sarif_files"\n```\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n### diff\n\n```\nusage: sarif diff [-h] [--output FILE] [--blame-filter FILE] old_file_or_dir new_file_or_dir\n\nFind the difference between two [sets of] SARIF files\n\npositional arguments:\n  old_file_or_dir       An old SARIF file or a directory containing the old SARIF files\n  new_file_or_dir       A new SARIF file or a directory containing the new SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n```\n\nPrint the difference between two [sets of] SARIF files.\n\nDifference between the issues in two SARIF files:\n```shell\nsarif diff "C:\\temp\\old_sarif_files\\devskim_myapp.sarif" "C:\\temp\\sarif_files\\devskim_myapp.sarif"\n```\n\nDifference between the issues in two directories of SARIF files:\n```shell\nsarif diff "C:\\temp\\old_sarif_files" "C:\\temp\\sarif_files"\n```\n\nWrite output to JSON file instead of printing to stdout:\n\n```shell\nsarif diff -o mydiff.json "C:\\temp\\old_sarif_files\\devskim_myapp.sarif" "C:\\temp\\sarif_files\\devskim_myapp.sarif"\n```\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n### html\n\n```\nusage: sarif html [-h] [--output PATH] [--blame-filter FILE] [--no-autotrim] [--image IMAGE] [--trim PREFIX] [file_or_dir [file_or_dir ...]]\n\nWrite an HTML representation of SARIF file(s) for viewing in a web browser\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output PATH, -o PATH\n                        Output file or directory\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n  --no-autotrim, -n     Do not strip off the common prefix of paths in the output document\n  --image IMAGE         Image to include at top of file - SARIF logo by default\n  --trim PREFIX         Prefix to strip from issue paths, e.g. the checkout directory on the build agent\n```\n\nCreate an HTML file summarising SARIF results.\n\n```shell\nsarif html -o summary.html "C:\\temp\\sarif_files"\n```\n\nUse the `--trim` option to strip specific prefixes from the paths, to make the generated HTML page less verbose.  The longest common prefix of the paths will be trimmed unless `--no-autotrim` is specified.\n\nUse the `--image` option to provide a header image for the top of the HTML page.  The image is embedded into the HTML, so the HTML document remains a portable standalone file.\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n### info\n\n```\nusage: sarif info [-h] [--output FILE] [file_or_dir [file_or_dir ...]]\n\nPrint information about SARIF file(s) structure\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n```\n\nPrint information about the structure of a SARIF file or multiple files.  This is about the JSON\nstructure rather than any meaning of the results produced by the tool.  The summary includes the\nfull path of the file, its size and modified date, the number of runs, and for each run, the\ntool that generated the run, the number of results, and the entries in the results\' property bags.\n\n```\nc:\\temp\\sarif_files\\ios_devskim_output.sarif\n  1256241 bytes (1.2 MiB)\n  modified: 2021-10-13 21:50:01.251544, accessed: 2022-01-09 18:23:00.060573, ctime: 2021-10-13 20:49:00\n  1 run\n    Tool: devskim\n    1323 results\n    All results have properties: tags, DevSkimSeverity\n```\n\n### ls\n\n```\nusage: sarif ls [-h] [--output FILE] [file_or_dir [file_or_dir ...]]\n\nList all SARIF files in the directories specified\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n```\n\nList SARIF files in one or more directories.\n\n```shell\nsarif ls "C:\\temp\\sarif_files" "C:\\temp\\sarif_with_date"\n```\n\n### summary\n\n```\nusage: sarif ls [-h] [--output FILE] [file_or_dir [file_or_dir ...]]\n\nList all SARIF files in the directories specified\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n```\n\nPrint a summary of the issues in one or more SARIF file(s), grouped by severity and then ordered by number of occurrences.\n\nWhen directories are provided as input and output, a summary is written for each input file, along with another file containing the totals.\n\n```shell\nsarif summary -o summaries "C:\\temp\\sarif_files"\n```\n\nWhen no output directory or file is specified, the overall summary is printed to the standard output.\n\n```shell\nsarif summary "C:\\temp\\sarif_files\\devskim_myapp.sarif"\n```\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n### trend\n\n```\nusage: sarif trend [-h] [--output FILE] [--blame-filter FILE] [--dateformat {dmy,mdy,ymd}] [file_or_dir [file_or_dir ...]]\n\nWrite a CSV file with time series data from SARIF files with "yyyymmddThhmmssZ" timestamps in their filenames\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n  --dateformat {dmy,mdy,ymd}, -f {dmy,mdy,ymd}\n                        Date component order to use in output CSV. Default is `dmy`\n```\n\nGenerate a CSV showing a timeline of issues from a set of SARIF files in a directory.  The SARIF file names must contain a\ntimestamp in the specific format `yyyymmddThhhmmss` e.g. `20211012T110000Z`.\n\nThe CSV can be loaded in Microsoft Excel for graphing and trend analysis.\n\n```shell\nsarif trend -o timeline.csv "C:\\temp\\sarif_with_date" --dateformat dmy\n```\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n### usage\n\n```\nusage: sarif usage [-h] [--output FILE]\n\n(Command optional) - print usage and exit\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FILE, -o FILE\n                        Output file\n```\n\nPrint usage and exit.\n\n### word\n\n```\nusage: sarif word [-h] [--output PATH] [--blame-filter FILE] [--no-autotrim] [--image IMAGE] [--trim PREFIX] [file_or_dir [file_or_dir ...]]\n\nProduce MS Word .docx summaries of the SARIF files specified\n\npositional arguments:\n  file_or_dir           A SARIF file or a directory containing SARIF files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output PATH, -o PATH\n                        Output file or directory\n  --blame-filter FILE, -b FILE\n                        Specify the blame filter file to apply. See README for format.\n  --no-autotrim, -n     Do not strip off the common prefix of paths in the output document\n  --image IMAGE         Image to include at top of file - SARIF logo by default\n  --trim PREFIX         Prefix to strip from issue paths, e.g. the checkout directory on the build agent\n```\n\nCreate Word documents representing a SARIF file or multiple SARIF files.\n\nIf directories are provided for the `-o` option and the input, then a Word document is produced for each individual SARIF file\nand for the full set of SARIF files.  Otherwise, a single Word document is created.\n\nCreate a Word document for each SARIF file and one for all of them together, in the `reports` directory (created if non-existent):\n```shell\nsarif word -o reports "C:\\temp\\sarif_files"\n```\n\nCreate a Word document for a single SARIF file:\n```shell\nsarif word -o "reports\\devskim_myapp.docx" "C:\\temp\\sarif_files\\devskim_myapp.sarif"\n```\n\nUse the `--trim` option to strip specific prefixes from the paths, to make the generated documents less verbose.  The longest common prefix of the paths will be trimmed unless `--no-autotrim` is specified.\n\nUse the `--image` option to provide a header image for the top of the Word document.\n\nSee [Blame filtering](blame-filtering) below for how to use the `--blame-filter` option.\n\n# Blame filtering\n\nUse the `sarif blame` command to augment a SARIF file or multiple SARIF files with blame information.\n\nBlame information is added to the property bag of each `result` object for which it was successfully obtained.  The keys and values used are as in the [git blame porcelain format](https://git-scm.com/docs/git-blame#_the_porcelain_format).  E.g.:\n\n```json\n{\n  "ruleId": "SM00702",\n  ...\n  "properties": {\n    "blame": {\n      "author": "aperson",\n      "author-mail": "<aperson@acompany.com>",\n      "author-time": "1350899798",\n      "author-tz": "+0000",\n      "committer": "aperson",\n      "committer-mail": "<aperson@acompany.com>",\n      "committer-time": "1350899798",\n      "committer-tz": "+0000",\n      "summary": "blah blah commit comment blah",\n      "boundary": true,\n      "filename": "src/net/myproject/mypackage/MyClass.java"\n    }\n  }\n}\n```\nNote that the bare `boundary` key is given the automatic value `true`.\n\nThis blame data can then be used for filtering and summarising via the `--blame-filter` option available for various commands.  This option requires a path to a filter-list file, containing a list of patterns and substrings to match against the blame information author email.  The format of a filter-list file is as follows:\n\n```\n# Lines beginning with # are interpreted as comments and ignored.\n# A line beginning with "description: " is interpreted as an optional description for the filter.  If no title is specified, the filter file name is used.\ndescription: Example filter from README.md\n# Lines beginning with "+: " are interpreted as inclusion substrings.  E.g. the following line includes issues whose author-mail field contains "@microsoft.com".\n+: @microsoft.com\n# The "+: " can be omitted.\n@microsoft.com\n# Instead of a substring, a regular expression can be used, enclosed in "/" characters.  Issues whose author-mail field includes a string matching the regular expression are included.  Use ^ and $ to match the whole author-mail field.\n+: /^<myname.*\\.com>$/\n# Again, the "+: " can be omitted for a regular expression include pattern.\n/^<myname.*\\.com>$/\n# Lines beginning with "-: " are interpreted as exclusion substrings.  E.g. the following line excludes issues whose author-mail field contains "bot@microsoft.com".\n-: bot@microsoft.com\n# Instead of a substring, a regular expression can be used, enclosed in "/" characters.  Issues whose author-mail field includes a string matching the regular expression are excluded.  Use ^ and $ to match the whole author-mail field.  E.g. the following pattern excludes all issues whose author-mail field contains a GUID.\n-: /[0-9A-F]{8}[-][0-9A-F]{4}[-][0-9A-F]{4}[-][0-9A-F]{4}[-][0-9A-F]{12}/\n```\n\nHere\'s an example of a filter-file that includes issues on lines changed by an `@microsoft.com` email address or a `myname.SOMETHING.com` email address, but not if those email addresses end in `bot@microsoft.com` or contain a GUID.  It\'s the same as the above example, with comments stripped out.\n\n```\ndescription: Example filter from README.md\n+: @microsoft.com\n+: /^<myname.*\\.com>$/\n-: bot@microsoft.com\n-: /[0-9A-F]{8}[-][0-9A-F]{4}[-][0-9A-F]{4}[-][0-9A-F]{4}[-][0-9A-F]{12}/\n```\n\nAll matching is case insensitive, because email addresses are.  Whitespace at the start and end of lines is ignored, which also means that line ending characters don\'t matter.  The blame filter file must be UTF-8 encoded (including plain ASCII7).  It can have a byte order mark or not.\n\nIf there are no inclusion patterns, all issues are included except for those matching the exclusion patterns.  If there are inclusion patterns, only issues matching the inclusion patterns are included.  If an issue matches one or more inclusion patterns and also at least one exclusion pattern, it is excluded.\n\nSometimes, there may be issues in the SARIF file to which the filter cannot be applied, because blame information is not available.  This can be for two reasons: either there is no blame information recorded for the file in which the issue occurred, or the issue location lacks a line number (or specifies line number 1 as a placeholder) so that blame information cannot be correlated to the issue.  These issues are included by default.  To identify which issues these are, create a filter file that excludes everything to which the filter can be applied:\n\n```\ndescription: Exclude everything filterable\n-: /.*/\n```\n\nThen run a `sarif` command using this filter file as the `--blame-filter` to see the default-included issues.\n\n# Usage as a Python library\n\nAlthough not its primary purpose, you can use sarif-tools from a Python script or module to\nload and summarise SARIF results.\n\n## Basic usage pattern\n\nAfter installation, use `sarif.loader` to load a SARIF file or files, and then use the operations\non the returned `SarifFile` or `SarifFileSet` objects to explore the data.\n\n```python\nfrom sarif import loader\n\nsarif_data = loader.load_sarif_file(path_to_sarif_file)\nissue_count_by_severity = sarif_data.get_result_count_by_severity()\nerror_histogram = sarif_data.get_issue_code_histogram("error")\n```\n\n## Result access API\n\nThe three classes defined in the `sarif_files` module, `SarifFileSet`, `SarifFile` and `SarifRun`,\nprovide similar APIs, which allows SARIF results to be handled similarly at multiple levels of\naggregation.  This section briefly describes some of the key APIs at the three levels of\naggregation.\n\n### get_distinct_tool_names()\n\nReturns a list of distinct tool names in a `SarifFile` or for all files in a `SarifFileSet`.\nA `SarifRun` has a single tool name so the equivalent method is `get_tool_name()`.\n\n### get_results()\n\nReturn the list of SARIF results.  These are objects as defined in the\n[SARIF standard section 3.27](https://docs.oasis-open.org/sarif/sarif/v2.1.0/os/sarif-v2.1.0-os.html#_Toc34317638).\n\n### get_records()\n\nReturn the list of SARIF results as simplified, flattened record dicts.  Each record has the\nattributes defined in `sarif_file.RECORD_ATTRIBUTES`.\n\n- `"Tool"` - the tool name for the run containing the result.\n- `"Severity"` - the SARIF severity for the record.  One of `error`, `warning` (the default if the\n  record doesn\'t specify) or `note`.\n- `"Code"` - the issue code from the result.\n- `"Location"` - the location of the issue, typically the file containing the issue.  Format varies\n  by tool.\n- `"Line"` - the line number in the file where the issue occurs.  Value is a string.  This defaults\n  to `"1"` if the tool failed to identify the line.\n\n### get_records_grouped_by_severity()\n\nAs per `get_records()`, but the result is a dict from SARIF severity level (`error`, `warning` and\n`note`) to the list of records of that severity level.\n\n### get_result_count(), get_result_count_by_severity()\n\nGet the total number of SARIF results.  `get_result_count_by_severity()` returns a dict from\nSARIF severity level (`error`, `warning` and `note`) to the integer number of results of that\nseverity.\n\n### get_issue_code_histogram(severity)\n\nFor the given severity, get histogram in the form of a list of pairs.  The first item in each pair\nis the issue code, the second item is the number of matching records, and the list is sorted in\ndecreasing order of frequency (the same as the `sarif summary` command output).\n\n### Disaggregation and filename access\n\nThese fields and methods allow access to the underlying information about the SARIF files.\n\n- `SarifFileSet.subdirs` - a list of `SarifFileSet` objects corresponding to the subdirectories of\n  the directory from which the `SarifFileSet` was created.\n- `SarifFileSet.files` - a list of `SarifFile` objects corresponding to the SARIF files contained\n  in the directory from which the `SarifFileSet` was created.\n- `SarifFile.get_abs_file_path()` - get the absolute path to the SARIF file.\n- `SarifFile.get_file_name()` - get the name of the SARIF file.\n- `SarifFile.get_file_name_without_extension()` - get the name of the SARIF file without its\n  extension.  Useful for constructing derived filenames.\n- `SarifFile.get_filename_timestamp()` - extract the timestamp from the filename of a SARIF file,\n  and return it as a string.  The timestamp must be in the format specified in the `sarif trend`\n  command.\n- `SarifFile.runs` - a list of `SarifRun` objects contained in the SARIF file.  Most SARIF files\n  only contain a single run, but it is possible to aggregate runs from multiple tools into a\n  single SARIF file.\n\n### Path shortening API\n\nCall `init_path_prefix_stripping(autotrim, path_prefixes)` on a `SarifFileSet`, `SarifFile` or `SarifRun` object to set up path filtering, either automatically removing the longest common prefix (`autotrim=True`) or removing specific prefixes (`autotrim=False` and a list of strings in `path_prefixes`).\n\n### Blame filtering API\n\nCall `init_blame_filter(filter_description, include_substrings, include_regexes, exclude_substrings, exclude_regexes)` on a `SarifFileSet`, `SarifFile` or `SarifRun` object to set up blame filtering.  `filter_description` is a string and the other parameters are lists of strings (with no `/` characters around the regular expressions).  They correspond in an obvious way to the filter file contents described in [Blame filtering](blame-filtering) above.\n\nCall `get_filter_stats()` to retrieve the filter stats after reading the results or records from sarif files.  It returns `None` if there is no filter, or otherwise a `sarif_file.FilterStats` object with integer fields `filtered_in_result_count`, `filtered_out_result_count`, `missing_blame_count` and `unconvincing_line_number_count`.  Call `to_string()` on the `FilterStats` object for a readable representation of these statistics, which also includes the filter file name or description (`filter_description` field).\n\n# Suggested usage in CI pipelines\n\nUsing the `--check` option in combination with the `summary` command causes sarif-tools to exit\nwith a nonzero exit code if there are any issues of the specified level, or higher.  This can\nbe useful to fail a continuous integration (CI) pipeline in the case of SAST violation.\n\nThe SARIF issue levels are `error`, `warning` and `note`.  These are all valid options for the\n`--check` option.\n\nE.g. to fail if there are any errors or warnings:\n\n```\nsarif --check warning summary c:\\temp\\sarif_files\n```\n\nThe `diff` command can check for any increase in issues of the specified level or above, relative\nto a previous or baseline build.\n\nE.g. to fail if there are any new issue codes at error level:\n\n```\nsarif --check error diff c:\\temp\\old_sarif_files c:\\temp\\sarif_files\n```\n\nYou can also use sarif-tools to filter and consolidate the output from multiple tools.  E.g.\n```\n# First run your static analysis tools, configured to write SARIF output.  How to do that depends\n# the tool.\n\n# Now run the blame command to augment the output with blame information.\nsarif blame -o with_blame/myapp_mytool_with_blame.sarif myapp_mytool.sarif\n\n# Now combine all tools\' output into a single file\nsarif copy --timestamp -o artifacts/myapp_alltools_with_blame.sarif\n```\n\nDownload the file `myapp_alltools_with_blame_TIMESTAMP.sarif` that is generated.  Then later you can\nfilter the results using the `--blame-filter` argument, or generate graph of code quality over time\nusing `sarif trend`.\n\n# Credits\n\nsarif-tools was originally developed during the Microsoft Global Hackathon 2021 by Simon Abykov, Nick Brabbs, Anthony Hayward, Sivaji Kondapalli, Matt Parkes and Kathryn Pentland.\n',
    'author': 'Microsoft',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/microsoft/sarif-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Code for `sarif diff` command.
"""

import json
import sys
from typing import Dict

from sarif.sarif_file import SarifFileSet, SARIF_SEVERITIES


def _occurrences(occurrence_count):
    return "1 occurence" if occurrence_count == 1 else f"{occurrence_count} occurrences"


def _signed_change(difference):
    return str(difference) if difference < 0 else f"+{difference}"


def print_diff(
    original_sarif: SarifFileSet, new_sarif: SarifFileSet, output, check_level=None
) -> str:
    """
    Generate a diff of the issues from the SARIF files and write it to stdout
    or a file if specified.
    original_sarif corresponds to the old files.
    new_sarif corresponds to the new files.
    """
    diff = calc_diff(original_sarif, new_sarif)
    if output:
        print("writing diff to", output)
        with open(output, "w", encoding="utf-8") as output_file:
            json.dump(diff, output_file, indent=4)
    else:
        for severity in SARIF_SEVERITIES:
            if diff[severity]["codes"]:
                print(
                    severity,
                    "level:",
                    _signed_change(diff[severity]["+"]),
                    _signed_change(-diff[severity]["-"]),
                )
                for (issue_code, old_count, new_count) in diff[severity]["codes"]:
                    if old_count == 0:
                        print(f'  New issue "{issue_code}" ({_occurrences(new_count)})')
                    elif new_count == 0:
                        print(f'  Eliminated issue "{issue_code}"')
                    else:
                        print(
                            f"  Number of occurrences {old_count} -> {new_count}",
                            f'({_signed_change(new_count - old_count)}) for issue "{issue_code}"',
                        )
            else:
                print(severity, "level: +0 -0 no changes")
        print(
            "all levels:",
            _signed_change(diff["all"]["+"]),
            _signed_change(-diff["all"]["-"]),
        )
    filter_stats = original_sarif.get_filter_stats()
    if filter_stats:
        print(f"  'Before' results were filtered by {filter_stats}")
    filter_stats = new_sarif.get_filter_stats()
    if filter_stats:
        print(f"  'After' results were filtered by {filter_stats}")
    ret = 0
    if check_level:
        for severity in SARIF_SEVERITIES:
            ret += diff.get(severity, {}).get("+", 0)
            if severity == check_level:
                break
    if ret > 0:
        sys.stderr.write(
            f"Check: exiting with return code {ret} due to increase in issues at or above {check_level} severity\n"
        )
    return ret


def calc_diff(original_sarif: SarifFileSet, new_sarif: SarifFileSet) -> Dict:
    """
    Generate a diff of the issues from the SARIF files.
    original_sarif corresponds to the old files.
    new_sarif corresponds to the new files.
    Return dict has keys "error", "warning", "note" and "all".
    """
    ret = {"all": {"+": 0, "-": 0}}
    for severity in SARIF_SEVERITIES:
        original_histogram = dict(original_sarif.get_issue_code_histogram(severity))
        new_histogram = new_sarif.get_issue_code_histogram(severity)
        new_histogram_dict = dict(new_histogram)
        ret[severity] = {"+": 0, "-": 0, "codes": []}
        if original_histogram != new_histogram_dict:
            for (issue_code, count) in new_histogram:
                old_count = original_histogram.pop(issue_code, 0)
                if old_count != count:
                    ret[severity]["codes"].append((issue_code, old_count, count))
                    if old_count == 0:
                        ret[severity]["+"] += 1
            for (issue_code, old_count) in original_histogram.items():
                ret[severity]["codes"].append((issue_code, old_count, 0))
                ret[severity]["-"] += 1
        ret["all"]["+"] += ret[severity]["+"]
        ret["all"]["-"] += ret[severity]["-"]
    return ret
